# -*- coding: utf-8 -*-
import base64
import logging
import io
import os
import requests
import os.path
import json
import warnings
from math import floor
from datetime import date
from dateutil.relativedelta import relativedelta
from collections import OrderedDict
from odoo import fields
from odoo.tools import html2plaintext

try:
    from ipwhois import IPWhois
except ImportError:
    pass

from odoo import SUPERUSER_ID as SU, _
from odoo.addons.saas_worker.config import saas_config
from odoo.addons.saas_worker.const import TEST, BUYAPPS_HOST, ODOO_SH_MASTER_URL, TIMEOUT
from odoo.addons.saas_worker.compat import db_monodb, route, request, http
from odoo.addons.saas_worker.util import from_role, recaptcha_v3_get_score, humanize_version
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.addons.auth_oauth.controllers.main import OAuthLogin
from odoo.addons.auth_signup.controllers.main import AuthSignupHome
from odoo.addons.loempia.util import CATEGORIES_APPS, CATEGORIES_THEMES, CATEGORIES_INDUSTRY, \
    FORBIDDEN_NETWORK_NAMES, ORDERS, RECENT_THRESHOLD, search_for_modules, sanitize_url
from odoo.tools import image_process, DEFAULT_SERVER_DATE_FORMAT
from odoo.osv import expression
from odoo.exceptions import UserError, ValidationError

from odoo.addons.web.controllers.home import Home


from werkzeug.exceptions import BadRequest, NotFound, Forbidden
from werkzeug.urls import url_encode


SEARCH_PAGE_SIZE = 20
STATS_LIMIT = 10
BLACKLIST_MODULE_FIELDS = [
    'repo_id',
    'series_id',
    'repo_maintainer_id',
    'repo_prio',
    'download_ids',
    'path',
    'ratings_ids',
    'product_id',
    'purchases',
    'score_details',
    'template_id',
    'temp_download_url',
]

logger = logging.getLogger(__name__)


def check_sec_fetch_dest(allowed_dests):
    """Verify the browser-provided Sec-Fetch-Dest, if provided
    in order to avoid serving a request when the destination
    context doesn't match. E.g. we should not return an image
    when the "Sec-Fetch-Dest" is `document`, to avoid accidentally
    returning an XSS shipped through a SVG file.

    Cfr: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-Fetch-Dest

    :param iterable allowed_dests: iterable list of allowed
        Sec-Fetch-Dest values (lowercase!)
    :raise: NotFound (404) if the Sec-Fetch-Dest header is present and
        not one of the allowed modes.
    """
    fetch_dest = request.httprequest.headers.get("Sec-Fetch-Dest")

    # if the header is missing, ignore the check. We can only
    # protect users who are using a supported browser, otherwise
    # we cannot determine the destination.
    if fetch_dest and fetch_dest.lower() not in allowed_dests:
        raise NotFound()

class Home(Home):
    @http.route('/', type='http', auth="none")
    def index(self, s_action=None, db=None, **kw):
        return request.redirect('/apps', code=301)


class LoempiaFrontendController(http.Controller):
    # NOTE: Embed route was removed as 'One-click install' has been deactivated for years.

    def _check_user_agent(self, request):
        # Temporary disable 1-click install for security reasons
        agent = request.httprequest.user_agent.string
        if agent and "Python-urllib" in agent:
            return request.not_found()

    @route([
        '/loempia/download/tmp/<string:module_name>/<string:validity_date>/<int:product_id>/<string:dl_hash>.zip',
        '/loempia/download/tmp/<string:module_name>/<string:validity_date>/<string:version>/<string:dl_hash>.zip',
    ], type='http', methods=['GET'], auth='public', website=True)
    def tmp_download(self, module_name, dl_hash=None, validity_date=None, product_id=None, version=None, **kwargs):
        self._check_user_agent(request)
        # * The second url is used for internal purpose and allow employees to use download
        #    a module when it hasn't been purchased already (i.e. porudcid is not set)
        #
        # * Temporary download link received by email or accessible on the sale order.
        #   This renders a pages with all the version you can download. The customer can
        #   then choose the version he desires to download
        if product_id:
            modules = request.env['loempia.module'].search([('name', '=', module_name), ('product_id', '=', product_id)])
        elif version:
            modules = request.env['loempia.module'].search([('name', '=', module_name), ('version', '=', version)])
        if not modules:
            return request.not_found()
        return request.render('loempia.module_download_links', {
            'modules': modules,
            'validity_date': validity_date,
            'dl_hash': dl_hash,
        })

    @route('/loempia/verify-download', type='http', auth="public", methods=['POST'], website=True)
    def can_access_download_link(self, **kwargs):
        self._check_user_agent(request)

        module_name = kwargs['module_name']
        version = kwargs['version']
        dl_hash = kwargs['dl_hash']
        validity_date = kwargs.get('validity_date')

        whitelisted_ips = request.env['ir.config_parameter'].sudo().get_param('loempia.recaptcha_whitelist_ips', '').split(',')
        if not TEST and request.httprequest.remote_addr not in whitelisted_ips:
            recaptcha_v3_token = kwargs.get('grc3_token')
            if recaptcha_v3_token is None:
                logger.error("Apps download access: recaptcha v3 token not sent from IP %s", request.httprequest.remote_addr)
                return request.render('loempia.expired_link')

            recaptcha_v3_score = recaptcha_v3_get_score(
                                    saas_config['apps']['recaptcha_v3_secretkey'],
                                    recaptcha_v3_token
                                )

            if recaptcha_v3_score != 'timeout' and (recaptcha_v3_score == False or recaptcha_v3_score < 0.5):
                logger.error("Apps download access: recaptcha v3 score %r for IP %s, redirect to home",
                             recaptcha_v3_score, request.httprequest.remote_addr)
                return request.redirect('/')

        return self._download(module_name, version=version, dl_hash=dl_hash, validity_date=validity_date)

    @route([
        '/loempia/download/<string:module_name>/<string:version>/<string:rand>.zip',
        '/loempia/download/<string:module_name>/<string:version>.zip',
    ], type='http', methods=['GET'], auth='public', website=True)
    def direct_download(self, module_name, version=None, **kwargs):
        # /!\ No CAPTCHA, but no download count. This undocumented fallback direct-download route
        # is present for customers who integrated the Apps download URL in their deployment
        # scripts prior to our introduction of recaptchav3. In practice it is not much of an issue
        # as the route is undocumented and leeches could also do the same manually if they really
        # wanted to download the whole set of free modules.
        return self._download(module_name, version=version, dl_hash=False, increment_download=False)

    def _download(self, module_name, version, dl_hash, validity_date=None, increment_download=True):
        module = request.env['loempia.module'].search([
            ('name', '=', module_name),
            '|',
            ('series_id.name', '=', version),
            ('version', '=', version)
        ], limit=1)

        #RAR:deactivated for the moment as people who bought a blocked module should still be able to download it ...
        #if module.status not in ['approved', 'invalid'] and not request.env['res.users'].has_group('loempia.group_manager'):
        #    logger.debug("Module is blocked. Rogue link or fake download ?")
        #    return request.not_found()

        repo_id = module.repo_id.sudo()
        if repo_id.is_server or repo_id.is_odoo_enterprise:
            logger.debug("Module is blocked. Rogue link or fake download ?")
            return request.not_found()

        if module and validity_date and dl_hash:
            product_id = module.product_id or module._get_commercial_series_module_id().product_id
            validity_date = fields.Date.from_string(validity_date)
            if (
                not module._verify_download_code(dl_hash, validity_date, product_id)
                or validity_date < date.today()
            ):
                return request.render('loempia.expired_link')
            else:
                # The point of the temporary link is to avoid checking if the user has
                # bought the module. The get_zip function below will check which
                # dependency of the module needs to be zipped together, based on the
                # purchases of the current user. In the case of the temporary download
                # link, we will create the zip as superuser, so all dependencies will
                # be zipped, because the superuser has virtually bought everything.
                module = module.sudo()

        elif not module or (module.price and module.download_hash != dl_hash):
            logger.debug("404 no mod/version")
            return request.not_found()

        elif module.price and not request.env['res.users'].has_group('loempia.group_manager'):
            if request.env.user == request.env.ref('base.public_user'):
                return request.redirect('/web/login?redirect=' + module.page_url)

            Purchases = request.env['loempia.module.purchase']
            if not Purchases.user_has_purchase(module):
                return request.redirect('/apps?e=3')

        return self._get_zip(module, increment_download)

    def _get_zip(self, module, increment_download):
        # get_zip needs to check if the user has purchased the module dependencies,
        # so the user in the environement need to be the requesting user, even though
        # a lot of the operations we will need to do in get_zip will require sudo
        # If in test mode, we return an empty zip file
        if TEST:
            filename = 'test_file.zip'
            content = b'PK\x05\x06' + b'\x00' * 18  # empty zip content
        else:
            filename = '%s-%s.zip' % (module.name, module.version)
            content = module._get_zip(zip_deps=True)

        if not content:
            logger.debug("404 no content")
            return request.not_found()

        if increment_download:
            try:
                # The library uses some deprecated calls that the author is going to fix
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=UserWarning)
                    remote_addr = request.httprequest.remote_addr
                    network_name = IPWhois(remote_addr).lookup_rdap()['network']['name']
                    module.sudo().youpie_one_more_download(
                        remote_addr=remote_addr,
                        fake=request.httprequest.method == 'HEAD' or any([n in network_name for n in FORBIDDEN_NETWORK_NAMES])
                    )
            except Exception:
                # Must never abort the download
                logger.warning('Failed to update download counter, ignoring', exc_info=True)

        headers = [
            ('Content-Type', 'application/zip'),
            ('Content-Length', len(content)),
            ('Content-Disposition', 'attachment; filename="' + filename + '"')
        ]

        return request.make_response(content, headers=headers)

    @from_role("paas_master")
    @route([
        '/loempia/sh/download/<string:module_name>/<string:version>/<string:dl_hash>.zip',
        '/loempia/sh/download/<string:module_name>/<string:version>/<string:validity_date>/<string:dl_hash>.zip',
    ], type='http', methods=['GET'], auth='public', website=True)
    def sh_download(self, module_name, version=None, dl_hash=None, validity_date=None, **kwargs):
        return self._download(module_name, version, dl_hash, validity_date)

    @route([
        '/loempia/download/odoo-sh/<string:module_name>/<string:version>',
    ], type='http', auth='public', website=True)
    def download_odoo_sh(self, module_name, version=None, **kwargs):
        """
        :param string modules_name: the technical name of the module to download
        :param string version: version of the module

        :return redirect to the Odoo.sh controller to commit an app in a branch of a repository
        :raise Forbidden when the module is paid and the user has not paid for it
        """
        Modules = request.env['loempia.module']
        Purchases = request.env['loempia.module.purchase']

        # Determine download URL
        # Free module: regular download URL
        # Paid module: temporary (one day) download url
        module = Modules.search([
            ('name', '=', module_name),
            '|', ('series_id.name', '=', version), ('version', '=', version)
        ], limit=1)
        if not module.price:
            download_url = '/loempia/sh/download/%(name)s/%(version)s/%(hm)s.zip?deps' % {
                'name': module.name,
                'version': module.version,
                'hm': module.sudo().download_hash
            }
        else:
            if request.env.user != request.env.ref('base.public_user') and Purchases.user_has_purchase(module):
                validity_date = date.today() + relativedelta(days=1)
                hm = module._get_download_code(validity_date=validity_date)
                download_url = '/loempia/sh/download/%(name)s/%(version)s/%(validity_date)s/%(hm)s.zip?deps' % {
                    'name': module.name,
                    'version': module.version,
                    'validity_date': validity_date,
                    'hm': hm,
                }
            else:
                raise Forbidden()

        # Redirect to Odoo.sh, with the download link, module name and icon.
        odoo_sh_download_link = '%s/app/add?%s' % (ODOO_SH_MASTER_URL, url_encode({
            'name': module.shortdesc,
            'author': module.author,
            'icon': '/apps/icon_image?module_id=%s' % module.id,
            'url': download_url,
            'version': module.series_id.name,
            'free': int(not module.price),
        }))
        return request.redirect(odoo_sh_download_link, local=False)

    @route([
        '/loempia/download/data_app/<string:module_name>/<string:version>',
    ], type='http', auth='public', website=True, methods=['GET'])
    def download_data_app(self, module_name, version=None, **kwargs):
        """
        :param string modules_name: the technical name of the module to download
        :param string version: version of the module

        :return zip file of the requested module
        :raise Forbidden when the module is paid and the user has not paid for it
        """
        Modules = request.env['loempia.module']

        # Determine download URL
        # Free module: regular download URL
        # Paid module: temporary (one day) download url
        if version:
            version = humanize_version(version)
        module = Modules.search([
            ('name', '=', module_name),
            ('is_data_module', '=', True),
            ('approved', '=', True),
            '|', ('series_id.name', '=', version), ('version', '=', version)
        ], limit=1)
        if not module.price:
            return self._download(module.name, version=module.version, dl_hash=module.sudo().download_hash, validity_date=None)
        else:
            raise Forbidden()

    @route([
        '/loempia/listdatamodules'
    ], type='json', auth='public', methods=['POST'], website=True, csrf=False)
    def get_data_modules(self, **kw):
        fields_to_read = (set(kw.get('module_fields')) & set(request.env['loempia.module']._fields)) - set(BLACKLIST_MODULE_FIELDS)
        fields_list = list(fields_to_read)
        if 'write_date' in fields_list:
            fields_list.remove('write_date')
        if not kw.get('module_name', False):
            category = ''
            search = False
            if kw.get('domain', False):
                for dom in kw.get('domain', []):
                    if dom[0] == 'category_id' and dom[1] == 'child_of':
                        category = dom[2]
                    if dom[0] == 'name' and dom[1] == 'ilike':
                        search = dom[2]
            post = {
                'order': 'Downloads',
                'data_module': True,
                'series': humanize_version(kw.get('series', '')),
                'search': search,
                'limit': kw.get('limit', None),
                'offset': kw.get('offset', 0),
            }
            modules_to_read = search_for_modules(
                request.env, kw.get('module_type', 'modules'),
                category,
                maintainer_id=None,
                **post,
            )
        else:
            domain = [
                ('name', '=', kw.get('module_name')),
                ('is_data_module', '=', True),
                ('approved', '=', True),
            ]
            if kw.get('series'):
                domain = expression.AND([domain, [('series_id.name', '=', kw.get('series'))]])
            modules_to_read = request.env['loempia.module'].sudo().search(
                domain,
                limit=kw.get('limit', None),
                offset=kw.get('offset', None),
            ).ids
        loempia_modules = request.env['loempia.module'].sudo().browse(modules_to_read).read(fields_list)
        return loempia_modules

    @route([
        '/loempia/listindustrycategory'
    ], type='json', auth='public', methods=['POST'], website=True, csrf=False)
    def get_industry_categories(self, **kw):
        categories = request.env['ir.module.category'].sudo().search_read([
            ('name', 'in', CATEGORIES_INDUSTRY),
            ('parent_id', '=', False),
        ], ['name', 'id'])
        return [{'id': cat['id'], 'display_name': cat['name'], '__count': 0} for cat in categories]

class AppsFrontendController(http.Controller):
    @route('/apps/charts', type='http', auth='public')
    def charts(self, serie):
        """return a list (in json) of l10n modules available on apps for the requested series"""
        category_id = request.env.ref('base.module_category_localization_account_charts').id
        M = request.env['loempia.module']
        modules = M.search([('series_id.name', '=', serie), ('category_id', '=', category_id),
                            ('approved', '=', True)], order='shortdesc')
        charts = [(m.name, m.shortdesc) for m in modules]

        content = json.dumps(charts)
        headers = [
            ('Content-Type', 'application/json'),
            ('Content-Length', len(content)),
        ]
        return request.make_response(content, headers)

    @route('/apps/login', type='http', auth='public', website=True, sitemap=False)
    def login(self):
        request.session.ensure_valid()

        DB = db_monodb(request)
        if not DB:
            raise RuntimeError('loempia module can only work in mono db matching mode. Please specify a --db-filter option that match only one database')

        logged = request.session._uid and request.session._dbname == DB

        if logged:
            return request.redirect('/web')

        with request.registry_cr() as (registry, cr):
            provider = registry['ir.model.data'].get_object(cr, SU, 'auth_oauth', 'provider_openerp')

            redirect_uri = request.httprequest.host_url + 'auth_oauth/signin'
            state = json.dumps(dict(d=DB, p=provider.id))
            params = dict(response_type='token',
                          client_id=provider.client_id,
                          redirect_uri=redirect_uri,
                          scope=provider.scope,
                          state=state,
                          )
            url = provider.auth_endpoint + '?' + url_encode(params)
            return request.redirect(url, local=False)

    def _get_user_pricelist(self):
        if request.session.get('website_apps_user_pricelist'):
            pricelist = request.env['loempia.pricelist'].browse(request.session['website_apps_user_pricelist'])
        else:
            country_code = request.geoip.get('country_code') or 'BE'
            pricelist = request.env['loempia.pricelist']._select_corresponding_pricelist(country_code)
            request.session['website_apps_user_pricelist'] = pricelist.id
        return pricelist

    @route([
        '/apps/<series>/<module_name>',
        '/apps/<any("modules", "themes", "industries"):module_type>/<series>/<module_name>',
        '/apps/assets/<series>/<module_name>/<path:file_path>',
        '/apps/assets/<any("modules", "themes", "industries"):module_type>/<series>/<module_name>/<path:file_path>',
    ], type='http', auth='public', website=True, strict_slashes=True)
    def apps_path(self, series, module_name, module_type=None, file_path="index.html", **post):
        """
        Route /assets/* is introduce to easily match the cdn filter instead of making some complex regex
        """
        pricelist = self._get_user_pricelist()

        dom = [('name', '=', module_name), ('series_id.name', '=', humanize_version(series))]
        if module_type == 'themes':
            dom += [('is_theme', '=', True)]
        elif module_type in ('modules', 'industries'):
            dom += [('is_theme', '=', False)]
        module = request.env['loempia.module'].with_context(pricelist_id=pricelist.id).sudo().search(dom)

        if not module:
            return request.not_found()

        module_type = module_type or ('themes' if module.is_theme else 'modules')
        root = os.path.abspath(os.path.join(module.path, 'static', 'description'))
        path = os.path.abspath(os.path.join(root, file_path))

        can_comment = True  # Users can now comment as many times as they want, but only the last rating is kept.

        # limit the access to files inside the app description directory
        if not path.startswith(root):
            raise BadRequest()
        if post.get('no_header', False):
            request.update_context(no_header=True)
        if not os.path.exists(path) and path.endswith("index.html"):
            return self.serve(module_type, path, module, can_comment, rst=True)

        if path.endswith(".html"):
            return self.serve(module_type, path, module, can_comment)
        elif os.path.exists(path):
            return self.serve_static(path)
        else:
            return request.not_found()


    # might become outdated -> use the link /apps/ apps_reference on odoo.com
    @route('/apps/product/<int:product_id>', type='http', auth='public', website=True, strict_slashes=True)
    def apps_from_product_id(self, product_id, **post):
        modules = request.env['loempia.module'].sudo().search([('product_id', '=', product_id)])
        if modules.exists():
            m = modules[0]._get_commercial_series_module_id()
            return request.redirect(m.page_url)
        else:
            return request.not_found()

    #########################
    #  Information Section  #
    #########################
    @route([
        '/apps/sales-conditions',
        '/apps/refund-policy'
    ], type='http', auth='public', website=True, sitemap=True)
    def sales_conditions(self, **post):
        reasons = request.env['loempia.refund.reason'].sudo().search([])
        return request.render(
            'loempia.sales_conditions',
            {
                'reasons': reasons,
                'max_days': request.env['ir.config_parameter'].sudo().get_param('loempia.refund_limit_days', 30)
            }
        )

    @route('/apps/repository-keys', type='http', auth='public', website=True, sitemap=True)
    def repo_keys(self, **post):
        home = os.path.expanduser("~")
        sshdir = os.path.join(home, ".ssh")
        values = {}
        bit_filename = os.path.join(sshdir, "id_rsa.pub")
        if os.path.isfile(bit_filename):
            with open(bit_filename) as f:
                values['generic_key'] = f.read()

        bit_filename = os.path.join(sshdir, "bitbucket_id_ed25519.pub")
        if os.path.isfile(bit_filename):
            with open(bit_filename) as f:
                values['bitbucket_key'] = f.read()

        return request.render('loempia.repo_keys', values)

    @route('/apps/vendor-guidelines', type='http', auth='public', website=True, sitemap=True)
    def vendor_guidelines(self, **post):
        return request.render('loempia.vendor_guidelines')

    @route('/apps/faq', type='http', auth='public', website=True, sitemap=True)
    def faq(self, **post):
        return request.render('loempia.faq')

    @route('/apps/upload', type='http', auth='public', website=True, sitemap=True)
    def upload_faq(self, **post):
        return request.render('loempia.upload_repo')

    #############################
    #  Repo Management section  #
    #############################
    @route('/apps/upload_repo', type='json', auth='user', website=True)
    def upload_repo(self, url, repo_id=None):
        if request.env.user.login == 'anonymous':
            return {'error': 'Please login with your personal account before uploading a repository'}
        if request.env.user.banned:
            return {'error': 'You cannot upload a repository during an active period of ban.'}

        # Test if the series is valid
        branch_name = url.split('#')[1]
        series_id = request.env['loempia.series'].search([('name', '=', branch_name)])
        if not series_id.exists():
            return {'error': 'Your branch must match exactly the name of the series your module is meant for. Ex: #8.0 for 8.0 series.'}
        try:
            url = sanitize_url(url)  # don't take care of 'bzr's vcs
        except Exception:
            return {'error': 'The provided url is invalid. Please consult our <a=href"/apps/faq#maintainer_faq_05a">FAQ</a>.'}

        same_url_repo = request.env['loempia.repo'].sudo().search([('url', '=', url)])
        # If repo_id is given, we are trying to change the url of an existing repo
        if repo_id:
            repo_id = request.env['loempia.repo'].search([('id', '=', repo_id)])
            # There is an error if repo_id is wrong, or when the url is already registered by another repo than repo_id itself
            if not repo_id.exists() or (same_url_repo.exists() and repo_id not in same_url_repo):
                return {'error': 'This repository and branch are already registered.'}
            old_url = repo_id.url
            module_branch = old_url.split('#')[1]
            if (repo_id.state == 'draft' or not repo_id.module_ids or module_branch == branch_name):
                if old_url != url:
                    migration_tool = request.env["loempia.repo.wizard.migrate"].sudo().create({
                        'repo_id': repo_id.id,
                        'url': url,
                    })
                    try:
                        migration_tool.migrate()
                    except ValidationError:
                        return {'error': 'This directory already exists. Please contact apps@odoo.com.'}
            else:
                return {'error': 'You can not change the branch of an active repository. Please create a new one.'}

        # If repo_id is not given, we are trying to create a new repo
        else:
            if same_url_repo.exists():
                return {'error': 'This repository and branch are already registered.'}
            try:
                repo_id = request.env['loempia.repo'].sudo().create({
                    'url': url,
                    'vcs': 'git',
                    'series_id': series_id.id,
                    'maintainer_id': request.uid,
                })
                repo_id.message_subscribe(partner_ids=[request.env.user.partner_id.id])
            except UserError as e:
                return {'error': str(e)}
        if not repo_id:  # At this point repo_id is a record, if there is no record, somehting went wrong
            return {'error': 'something went wrong'}

        return dict(repo_id=repo_id.id, state=repo_id.state, scan_log=repo_id.scan_log)

    @route('/apps/repo_auto_scan', type='json', auth='user', website=True)
    def repo_auto_scan(self, repo_id, auto_scan):
        repo_id = request.env['loempia.repo'].search([('id', '=', repo_id), ('maintainer_id', '=', request.env.user.id)])
        if repo_id and repo_id.maintainer_id == request.env.user:
            repo_id.sudo().scan = auto_scan

    @route('/apps/scan_repo', type='json', auth='user', website=True)
    def force_scan_repo(self, repo_id):
        repo_id = request.env['loempia.repo'].search([('id', '=', repo_id), ('maintainer_id', '=', request.env.user.id)])
        logger.info('scan of repo %s requested by %s - %s', repo_id.sudo().url, request.env.user.login, request.env.user.id)
        if repo_id:
            repo_id.sudo()._scan_repo('scan', force_scan=True)
            # On successfull upload, an app vendor is tagged as such
            if repo_id.state == "active":
                request.env.user.partner_id.write({'category_id': [(4, request.env.ref('loempia.dev_partner_tag').id)]})
            return dict(repo_id=repo_id.id, state=repo_id.state, scan_log=repo_id.scan_log)
        return dict(error='A repository with this ID could not be found.')

    @route('/apps/publish_repo', type='json', auth='user', website=True)
    def publish_repo(self, repo_id):
        repo_id = request.env['loempia.repo'].search([('id', '=', repo_id), ('maintainer_id', '=', request.env.user.id)])
        if repo_id:
            repo_id.sudo()._publish_repo()

    @route('/apps/unpublish_repo', type='json', auth='user', website=True)
    def unpublish_repo(self, repo_id):
        repo_id = request.env['loempia.repo'].search([('id', '=', repo_id), ('maintainer_id', '=', request.env.user.id)])
        if repo_id:
            repo_id.sudo()._unpublish_repo()

    @route('/apps/toggle_module_publish', type='json', auth='user', website=True)
    def toggle_module_publish(self, module_id):
        module = request.env['loempia.module'].browse(module_id)
        if module.exists() and module.repo_maintainer_id == request.env.user:
            module.sudo()._toggle_publish()

    #######################
    #  Dashboard section  #
    #######################
    @route([
        '/apps/contributors',
        '/apps/contributors/<string:module_type>',
        '/apps/contributors/<string:module_type>/page/<int:page>',
    ], type='http', auth='public', website=True, sitemap=True)
    def contributors(self, module_type='modules', page=1, by=50, **post):
        if module_type not in ['modules', 'themes']:
            return request.not_found()

        exclude = ['Unknown']

        orders = {
            'Contributions': 'contributions',
            'Rating': 'rating',  # I don't know why but using DESC gives ASC result ...
        }

        order = orders.get(post.get('order'), orders['Contributions'])

        args = {'srch': '%' + post.get('search', '') + '%'}
        request.cr.execute("""
            SELECT
                MIN(author),
                iauthor,
                COUNT(iauthor) AS contributions,
                TO_CHAR(MIN(create_date), 'DD Month YYYY') AS since,
                AVG(NULLIF(average_rating, 0)) AS rating
            FROM (
                SELECT
                    UNNEST(regexp_split_to_array(m.author, ', ')) AS author,
                    UNNEST(regexp_split_to_array(LOWER(m.author), ', ')) AS iauthor,
                    m.create_date,
                    t.average_rating
                FROM loempia_module_template as t
                JOIN loempia_module as m on t.latest_module_id = m.id
                -- TODO: Themes subcategories ?
                WHERE {theme} m.is_theme
                AND not m.is_hidden
                AND (m.status = 'approved' OR m.status= 'invalid')
                AND m.is_published
                AND not m.is_deleted
                AND (
                    m.summary ILIKE %(srch)s
                    OR m.shortdesc ILIKE %(srch)s
                    OR m.name ILIKE %(srch)s
                    OR m.description ILIKE %(srch)s
                )
            ) AS c
            GROUP BY iauthor
            ORDER BY {order} DESC NULLS LAST;
        """.format(theme='not' if module_type == 'modules' else '', order=order), args)

        res = request.cr.fetchall()
        authors = []
        contributions = {}
        dates = {}
        ratings = {}
        for a in res[by*(page-1):by*(page-1)+by]:
            author = a[0]
            if author in exclude:
                continue
            authors.append(author)
            contributions[author] = a[2]
            dates[author] = a[3]
            ratings[author] = a[4]

        nbr_pages = max((int(len(res) / by)) + (1 if len(res) % by > 0 else 0), 1)
        page = min(page, nbr_pages)
        pager = request.website.pager(
            url='/apps/contributors/' + module_type, total=nbr_pages, page=page, step=1,
            scope=5, url_args=post
        )

        return request.render('loempia.contributors', {
            'orders': orders,
            'module_type': module_type,
            'authors': authors,
            'contributions': contributions,
            'dates': dates,
            'ratings': ratings,
            'pager': pager,
            'post': post,
            'need_navbar': True,
        })

    @route([
        '/apps/dashboard',
        '/apps/dashboard/stats',
    ], type='http', auth="user", website=True)
    def dashboard(self, **post):
        values = self._prepare_stats_data(user_id=request.env.uid)
        values.update(stats=True)
        return request.render('loempia.dashboard_stats', values)

    @route([
        '/apps/dashboard/repos',
        '/apps/dashboard/repos/page/<int:page>',
    ], type='http', auth="user", website=True)
    def dashboard_repos(self, page=1, search=None, search_in='all', **post):

        search_inputs = {
            'all': {'input': 'all', 'label': 'All'},
            'url': {'input': 'url', 'label': 'Url'},
            'series': {'input': 'series', 'label': 'Series'},
        }

        domain = [('maintainer_id', '=', request.env.uid)]

        if search and search_in:
            search_domain = []
            if search_in in ('url', 'all'):
                search_domain = expression.OR([search_domain, [('url', 'ilike', search)]])
            if search_in in ('series', 'all'):
                search_domain = expression.OR([search_domain, [('series_id.name', 'ilike', search)]])
            domain = expression.AND([domain, search_domain])

        repo_count = request.env['loempia.repo'].search_count(domain)

        url_args = dict(post, search=search, search_in=search_in)
        pager = request.website.pager(
            url='/apps/dashboard/repos', total=repo_count, page=page, step=SEARCH_PAGE_SIZE,
            scope=5, url_args= url_args,
        )
        pager_last = request.website.pager(
            url='/apps/dashboard/repos', total=repo_count, page=page, step=SEARCH_PAGE_SIZE,
            scope=10000, url_args= url_args,
        )

        repos = request.env['loempia.repo'].search(domain, order='series_id,url', limit=SEARCH_PAGE_SIZE, offset=pager['offset'])

        return request.render('loempia.dashboard_repos', {
            'pager': pager,
            'pager_last': pager_last['page_end'],
            'search': search,
            'search_inputs': OrderedDict(sorted(search_inputs.items())),
            'repo_ids': repos,
        })

    @route([
        '/apps/dashboard/preferences',
    ], type='http', auth="user", website=True)
    def dashboard_prefs(self, **post):
        return request.render('loempia.dashboard_prefs', {
            'partner': request.env.user.partner_id,
            'countries': request.env['res.country'].sudo().search([]),
            'submitted': post.get('submitted', False),
        })

    ######################
    #  Comments section  #
    ######################
    @route('/apps/report-rating', type='json', auth="user", website=True)
    def report_rating(self, rating_id=0, **post):
        rating = request.env['loempia.module.rating'].sudo().browse(int(rating_id))
        user = request.env.user
        body = html2plaintext(rating.body)
        response = False
        if rating.exists():
            if rating.state == 'pending':
                rating.write({'report_count': rating.report_count + 1})
                response = True
            elif rating.state == 'unreviewed':
                reason = post.get('reason')
                response = rating._report(reason, user, body)

        return response

    @route('/apps/rating', type='http', auth="user", website=True)
    def post_rating(self, module_id=0, **post):
        url = '/apps'
        try:
            module_id = int(module_id)
            user = request.env.user
            m = request.env['loempia.module'].browse(module_id).sudo()
            rating = int(post.get('rating') or 0)
            rating = min(max(rating, 0), 5)

            ratings = m.ratings_ids.filtered(lambda r: r.author_id == user.partner_id)
            if m.exists():
                Purchases = request.env['loempia.module.purchase']
                if not post.get('force_post') and (rating in [1, 2]) and m.price != 0.0 and Purchases.user_has_purchase(m):
                    return LoempiaDisputesController().fuzzy_support(
                        module_id=m.id,
                        rating=post['rating'],
                        subject=post['subject'],
                        comment=post['comment']
                    )
                url = "{module.page_url}/#loempia-comments".format(module=m)

                if not Purchases.user_has_purchase(m):
                    rating = 0  # Not a rating, just a comment

                message = m.ratings_ids.create({
                    'module_id': module_id,
                    'template_id': m.template_id.id,
                    'rating': rating,
                    'subject': post['subject'],
                    'body': post['comment'],
                    'author_id': user.partner_id.id,
                })

                # Instead of allowing only one rating, we invalidate the previous ones
                # This removes the need for an edit function, and allows people to use
                # the ratings as a mean to contact the developer and chat about features,
                # support etc, so developers can now respond to criticism in the comments.
                if message and rating:
                    ratings.write({'rating': 0})

                # If it is a paid module, mark the corresponding PO as 'rated' in case of an actual rating (>0)
                if Purchases.user_has_purchase(m) and rating > 0:
                    states = ['confirmed', 'done', 'redeemed']
                    commercial_version = m._get_commercial_series_module_id()
                    p = Purchases.search([
                        '|',
                        ('product_id', '=', commercial_version.product_id),
                        ('module_id', '=', commercial_version.id),
                        ('user_id', '=', request.env.user.id),
                        ('state', 'in', states),
                    ])
                    p.sudo().write({'rated': True})

                # Notify the Developer that a comment was posted
                message._module_notification(m.repo_id.maintainer_id.partner_id, reply=False)

        except (ValueError, KeyError):
            pass
        return request.redirect(url)

    @route('/apps/reply', type='http', auth="user", website=True)
    def post_reply(self, parent_id=0, **post):
        url = '/apps'
        try:
            parent_id = int(parent_id)
            user = request.env.user
            rating = request.env['loempia.module.rating'].browse(parent_id).sudo()
            if rating.exists():
                m = rating.module_id
                # Only the maintainer is allowed to reply
                if m.repo_id.maintainer_id != user:
                    return request.redirect(url)

                url = "{module.page_url}/#loempia-comments".format(module=m)

                request.update_context(mail_create_nosubscribe=True)
                message = rating.create({
                    'module_id': m.id,
                    'template_id': m.template_id.id,
                    'rating': 0,
                    'subject': post['subject'],
                    'body': post['comment'],
                    'author_id': user.partner_id.id,
                    'parent_id': parent_id,
                })
                # Notify the orignal author the developer answered
                message._module_notification(message.parent_id.author_id, reply=True)
        except (ValueError, KeyError):
            pass
        return request.redirect(url)

    ######################
    # Odoo Apps Homepage #
    ######################
    # For some reason, the route '/apps/<any("modules", "themes"):module_type>' doesn't work ...
    @route([
        '/apps',
        '/apps/modules',
        '/apps/modules/category/<string:category>',
    ], type='http', auth='public', website=True, sitemap=True)
    def apps_welcome_modules(self, category='', **post):
        if category not in CATEGORIES_APPS:
            category = ''
        return self.apps_welcome(module_type='modules', category=category, **post)

    @route([
        '/apps/themes',
        '/apps/themes/category/<string:category>',
    ], type='http', auth='public', website=True, sitemap=True)
    def apps_welcome_themes(self, category='', **post):
        if category not in CATEGORIES_THEMES:
            category = ''
        return self.apps_welcome(module_type='themes', category=category, **post)

    @route([
        '/apps/industries',
        '/apps/industries/category/<string:category>',
    ], type='http', auth='public', website=True)
    def apps_welcome_industries(self, category='', **post):
        if category not in CATEGORIES_INDUSTRY:
            category = ''
        return self.apps_welcome(module_type='industries', category=category, **post)

    def apps_welcome(self, module_type='modules', category='', **post):
        pricelist = self._get_user_pricelist()
        user_id = request.env.user.id
        env = request.env

        best_sellers_ids = search_for_modules(env, module_type, category, user_id, limit=4, **dict(post, order='Relevance'))
        newests_ids = search_for_modules(env, module_type, category, user_id, limit=4, **dict(post, order='Newest'))
        most_downloaded_ids = search_for_modules(env, module_type, category, user_id, limit=4, **dict(post, order='Downloads'))

        Modules = request.env['loempia.module'].with_context(pricelist_id=pricelist.id).sudo()

        Series = request.env['loempia.series']
        all_series = Series.sudo().search([('name', 'not ilike', 'saas-%')])
        default_series = all_series[0].name
        categories = CATEGORIES_APPS
        if module_type == 'themes':
            categories = CATEGORIES_THEMES
        elif module_type == 'industries':
            categories = CATEGORIES_INDUSTRY

        return request.render('loempia.apps_welcome', {
            'best_seller_ids': Modules.browse(best_sellers_ids),
            'newest_ids': Modules.browse(newests_ids),
            'most_downloaded_ids': Modules.browse(most_downloaded_ids),
            "series": all_series,
            "default_series": default_series,
            "categories": categories,
            "category": category,
            "orders": ORDERS,
            'post': post,
            'module_type': module_type,
            'currency_id': pricelist.currency_id,
            'need_subnav': True,
            'need_navbar': True,
            'is_welcome_page': True,
        })

    ######################
    # Module Search Page #
    ######################
    @route([
        '/apps/browse',
        '/apps/<any("modules", "themes", "industries"):module_type>/browse',
        '/apps/<any("modules", "themes", "industries"):module_type>/category/<string:category>/browse',
        '/apps/<any("modules", "themes", "industries"):module_type>/browse/page/<int:page>',
        '/apps/<any("modules", "themes", "industries"):module_type>/category/<string:category>/browse/page/<int:page>',
    ], type='http', auth='public', website=True, sitemap=True)
    def module_search(self, module_type='modules', category="", page=1, **post):
        user_id = request.env.user.id
        pricelist = self._get_user_pricelist()
        categories = CATEGORIES_APPS
        if module_type == 'themes':
            categories = CATEGORIES_THEMES
        elif module_type == 'industries':
            categories = CATEGORIES_INDUSTRY
        default_series = ''
        for i in list(post):
            if not post[i]:
                del post[i]

        Series = request.env['loempia.series'].sudo()
        env = request.env

        all_series = Series.search([('name', 'not ilike', 'saas-%')])
        default_series = all_series[0].name

        search_modules = search_for_modules(env, module_type=module_type, category=category, maintainer_id=user_id, **post)
        count = len(search_modules)

        # Displaying results according to current page
        nbr_pages = max((int(count / SEARCH_PAGE_SIZE)) + (1 if count % SEARCH_PAGE_SIZE > 0 else 0), 1)
        page = min(int(page), nbr_pages)

        offset = (page - 1) * SEARCH_PAGE_SIZE
        limit = offset + SEARCH_PAGE_SIZE

        modules = env['loempia.module'].sudo().with_context(pricelist_id=pricelist.id).browse(search_modules[offset:limit])

        pager = request.website.pager(
            url='/apps/' + module_type + (('/category/' + category) if category else '') + '/browse', total=nbr_pages, page=page, step=1,
            scope=5, url_args=post
        )

        pager_last = request.website.pager(
            url='/apps/' + module_type + (('/category/' + category) if category else '') + '/browse', total=nbr_pages, page=page, step=1,
            scope=10000, url_args=post
        )

        category_id = request.env['ir.module.category'].sudo().search([('name', '=ilike', category)])

        # Rendering the web page
        return request.render('loempia.module_search', {
            "series": all_series,
            "default_series": default_series,
            "categories": categories,
            "category": category,
            "main_object": category_id[0] if category_id.exists() else False,
            "modules": modules,
            "page": page,
            "nbr_pages": nbr_pages,
            "orders": ORDERS,
            "count": count,
            "pager": pager,
            "pager_last": pager_last['page_end'],
            'post': post,
            'module_type': module_type,
            'currency_id': pricelist.currency_id,
            'need_subnav': True,
            'need_navbar': True,
        })

    @http.route(['/apps/preferences'], type='http', auth='user', website=True)
    def prefs(self, **post):
        if request.env.user.login == 'anonymous':
            raise Forbidden()

        partner = request.env['res.users'].browse(request.uid).partner_id
        if post:
            notify_on_module_comment = post.pop('notify_on_module_comment', "False") == "True"
            result = partner.sudo().write({'notify_on_module_comment': notify_on_module_comment})
            post.update({'submitted': result})

        return self.dashboard_prefs(**post)

    @route([
        '/apps/statistics',
        '/apps/<any("modules", "themes", "industries"):module_type>/statistics',
        '/apps/<any("modules", "themes", "industries"):module_type>/category/<string:category_name>/statistics',
    ], type='http', auth='public', website=True, sitemap=True)
    def apps_stats(self, module_type="modules", category_name=None, user_id=None, **post):
        return request.render('loempia.statistics', self._prepare_stats_data(module_type, category_name, user_id))

    @route('/apps/statistics/graphs', type='json', auth='public')
    def apps_graph_stats(self, module_type="modules", category_id=None, user_id=None, **post):
        if user_id:
            try:
                uid = int(user_id)
                # request.session.uid != request.env.uid (empty with json)
                if uid != request.session.uid:
                    raise ValueError
                user_id = uid
            except ValueError:
                user_id = None

        category_id = category_id or None
        if category_id:
            try:
                category_id = int(category_id)
                if not request.env['ir.module.category'].browse(category_id).exists():
                    raise ValueError
            except ValueError:
                category_id = None

        stats = self._get_stats(module_type, category_id, user_id)
        top_trends = self._get_trends(module_type, category_id, user_id)

        # top_modules_graph = []
        top_contributors_graph = []
        top_sellers_graph = []
        top_trends_graphs = []
        all_sales_graph = []
        cumulative_sales_graphs = []

        results = []

        # Contributors
        for contr in stats['top_contributors']:
            top_contributors_graph.append([contr[0], contr[2]])
        results.append({
            'name': 'Downloads',
            'key': 'contributors',
            'values': top_contributors_graph,
        })

        # Sellers
        for seller in stats['top_sellers']:
            top_sellers_graph.append([seller[0], seller[2]])
        results.append({
            'name': 'Sales (€)',
            'key': 'sellers',
            'values': top_sellers_graph,
        })

        # TODO: barchart for module is not displayed (o_graph_module removed from xml)
        # Modules
        # for module in stats['top_modules']:
        #     top_modules_graph.append([module.shortdesc, module.sales_recent, module.name])
        # results.append({
        #     'name': 'Modules',
        #     'key': 'modules',
        #     'values': top_modules_graph,
        # })

        # Trends
        trends_modules = stats['top_modules'] if user_id else top_trends['modules']
        for module in trends_modules:
            # One line chart by module
            top_trends_graphs.append({
                'key': module.shortdesc,
                'values': self._build_trend_graph(module, cumulative=False),
            })
        results.append({
            'name': 'Sales (€)',
            'key': 'trends',
            'values': top_trends_graphs,
        })

        # Cumulative Sales
        for module in stats['top_modules']:
            # One line chart by module
            cumulative_sales_graphs.append({
                'key': module.shortdesc,
                'values': self._build_trend_graph(module, cumulative=True),
            })
        results.append({
            'name': 'Cumulative Sales (€)',
            'key': 'cumulative_sales',
            'values': cumulative_sales_graphs,
        })

        # All Sales
        all_sales_graph = self._build_sales_monthly_graph(module_type, category_id, user_id)
        results.append({
            'name': 'Sales (€)',
            'key': 'all_sales',
            'values': all_sales_graph,
        })

        # All Downloads
        # TODO

        return results

    def _prepare_stats_data(self, module_type="modules", category_name=None, user_id=None):
        category = None
        if category_name:
            category = request.env['ir.module.category'].sudo().search([('name', 'ilike', category_name)], limit=1)

        if user_id:
            try:
                uid = int(user_id)
                if uid != request.env.uid:
                    raise ValueError
                user_id = uid
            except ValueError:
                user_id = None

        # RAR - There is no point in calling the get_stat function here. This will be worked out in our refactoring
        # TODO: remove the useless calls
        # stats = self._get_stats(module_type, category and category.id or None, user_id)

        # TODO: kpi_best_sellers is no longer showed on the dashboard, no point to call it for now.
        # TODO: Add more KPIs ?
        # kpis = self._get_statistics_kpis()
        # kpi_best_sellers = kpis['sellers']

        top_trends = self._get_trends(module_type, category and category.id or None, user_id)
        return {
            'module_type': module_type,
            'category': category,
            'user': request.env['res.users'].sudo().browse(user_id).exists(),
            'top_trends': top_trends,
            # 'top_modules': stats['top_modules'],
            # 'top_contributors': stats['top_contributors'],
            # 'top_sellers': stats['top_sellers'],
            # 'kpi_best_sellers': kpi_best_sellers,
        }

    def _get_stats(self, module_type, category_id=None, user_id=None):
        # Top Modules
        is_theme = module_type == 'themes'
        # We need to add m.status = 'approved' because if all series are blocked we do not want them to appear in the stats
        request.cr.execute("""
            SELECT m.id
            FROM loempia_module m
            JOIN loempia_module_template t ON t.latest_module_id =m.id
            WHERE t.sales != 0
                AND (m.status = 'approved' OR m.status = 'invalid')
                AND NOT m.is_deleted
                AND m.is_theme = %(is_theme)s
                AND ((%(category_id)s IS NULL) OR (m.category_id = %(category_id)s))
                AND ((%(repo_maintainer_id)s IS NULL) OR (m.repo_maintainer_id = %(repo_maintainer_id)s))
            ORDER BY t.sales_recent DESC
            """, {
            'is_theme': is_theme,
            'category_id': category_id or None,
            'repo_maintainer_id': user_id or None
        })
        all_module_ids = request.cr.fetchall()
        all_modules = request.env['loempia.module'].sudo().browse([x[0] for x in all_module_ids])
        top_modules = all_modules[:STATS_LIMIT]

        # Hide top contributors & top sellers if user_id is set
        top_contributors = []
        top_sellers = []

        if not user_id:
            # If user_id, the user is the only seller/contributor
            # Top contributors
            request.cr.execute("""
                SELECT
                    MIN(author),
                    iauthor,
                    SUM(module_download) AS module_sum,
                    AVG(NULLIF(average_rating, 0)) AS rating
                FROM (
                    SELECT
                        UNNEST(regexp_split_to_array(m.author, ', ')) AS author,
                        UNNEST(regexp_split_to_array(LOWER(m.author), ', ')) AS iauthor,
                        t.download_count_all_recent as module_download,
                        t.average_rating as average_rating
                    FROM loempia_module_template as t
                    JOIN loempia_module as m on t.latest_module_id = m.id
                    WHERE
                        m.is_theme = %(is_theme)s AND
                        (%(category_id)s IS NULL) OR (m.category_id = %(category_id)s)
                        AND not (m.is_hidden OR m.is_deleted)
                        AND (m.status = 'approved' OR m.status = 'invalid')
                ) AS c
                WHERE module_download > 0
                GROUP BY iauthor
                ORDER BY module_sum DESC NULLS LAST;
            """, {
                'is_theme': is_theme,
                'category_id': category_id or None
            })
            top_contributors = request.cr.fetchall()
            top_contributors = top_contributors[:STATS_LIMIT]

            # Top sellers
            request.cr.execute("""
                SELECT
                    MIN(author),
                    iauthor,
                    SUM(module_sales) AS module_sum,
                    AVG(NULLIF(average_rating, 0)) AS rating
                FROM (
                    SELECT
                        UNNEST(regexp_split_to_array(m.author, ', ')) AS author,
                        UNNEST(regexp_split_to_array(LOWER(m.author), ', ')) AS iauthor,
                        t.sales_recent as module_sales,
                        t.average_rating as average_rating
                    FROM loempia_module_template as t
                    JOIN loempia_module as m on t.latest_module_id = m.id
                    WHERE
                        m.is_theme = %(is_theme)s AND
                        (%(category_id)s IS NULL) OR (m.category_id = %(category_id)s)
                        AND NOT (m.is_hidden or m.is_deleted)
                        AND (m.status = 'approved' OR m.status = 'invalid')
                ) AS c
                WHERE module_sales > 0
                GROUP BY iauthor
                ORDER BY module_sum DESC NULLS LAST;
            """, {
                'is_theme': is_theme,
                'category_id': category_id or None
            })
            top_sellers = request.cr.fetchall()
            top_sellers = top_sellers[:STATS_LIMIT]

        return {
            'all_modules': all_modules,
            'top_modules': top_modules,
            'top_contributors': top_contributors,
            'top_sellers': top_sellers,
        }

    def _get_trends(self, module_type, category_id=None, user_id=None):
        # Top trends
        # Trends metric : purchase_{last_month} / purchase_{last_3_month}
        # Example:
        #   last month: 100, last 2 months before: 100
        #   trend: (2*(100/100) - 1)*100 = +100%
        request.cr.execute("""
            SELECT p0.template_id, 100*(2*(p0.price / p1.price)-1) AS trend, p0.repo_maintainer_id, p0.category_id
            FROM
                (
                SELECT m.template_id AS template_id,SUM(p.price) AS price,
                m.repo_maintainer_id,array_agg(distinct(m.category_id)) as category_id
                FROM
                    loempia_module_purchase AS p,
                    loempia_module AS m
                WHERE
                    p.referrer_module_id = m.id AND
                    p.state not in ('draft', 'cancel') AND
                    p.price > 0 AND
                    p.date_order BETWEEN current_date - interval '%(recent)s days' AND current_date AND
                    m.is_theme = %(is_theme)s AND
                    (m.status = 'approved' OR m.status = 'invalid') AND
                    NOT m.is_deleted
                GROUP BY
                    m.template_id,m.repo_maintainer_id
                ) AS p0,
                (
                SELECT m.template_id AS template_id,SUM(p.price) AS price
                FROM
                    loempia_module_purchase AS p,
                    loempia_module AS m
                WHERE
                    p.referrer_module_id = m.id AND
                    p.state not in ('draft', 'cancel') AND
                    p.price > 0 AND
                    p.date_order BETWEEN current_date - interval '%(old)s days' AND current_date - interval '%(recent)s days' AND
                    m.is_theme = %(is_theme)s AND
                    (m.status = 'approved' OR m.status = 'invalid') AND
                    NOT m.is_deleted
                GROUP BY
                    m.template_id,m.repo_maintainer_id
                ) AS p1
            WHERE
                p0.template_id = p1.template_id
            ORDER BY
                trend DESC;
        """, {
            'is_theme': module_type == 'themes',
            'recent': RECENT_THRESHOLD,
            'old': 3 * RECENT_THRESHOLD
        })
        top_trends_infos = request.cr.fetchall()

        if user_id:
            # Only keep modules maintainted by the user
            top_trends_infos = [x for x in top_trends_infos if x[2] == user_id]

        if category_id:
            # Only keep modules in the good category
            top_trends_infos = [x for x in top_trends_infos if category_id in x[3]]

        # TODO: sort by created date before ?
        top_trends_infos = top_trends_infos[:STATS_LIMIT]

        trends_values = {}
        for el in top_trends_infos:
            trends_values[el[0]] = round(el[1], 2)

        top_trends = request.env['loempia.module.template'].sudo().browse([x[0] for x in top_trends_infos]).mapped('latest_module_id')

        return {
            'modules': top_trends,
            'values': trends_values,
        }

    def _get_statistics_kpis(self):
        #probably broken but unused atm (04-2020)
        # TODO if we use this kpi again, we must update the request with the status of the module
        # Top sellers
        request.cr.execute("""
            SELECT
                MIN(author),
                iauthor,
                SUM(module_sales) AS module_sum
            FROM (
                SELECT
                    UNNEST(regexp_split_to_array(m.author, ', ')) AS author,
                    UNNEST(regexp_split_to_array(LOWER(m.author), ', ')) AS iauthor,
                    t.sales_recent AS module_sales
                FROM loempia_module_template as t
                JOIN loempia_module as m on t.latest_module_id = m.id
            ) AS c
            GROUP BY iauthor
            ORDER BY module_sum DESC NULLS LAST;
        """)

        SELLERS_LIMIT = 20
        top_sellers = request.cr.fetchall()
        top_sellers = top_sellers[:SELLERS_LIMIT]
        kpi_top_sellers = int(sum(x[2] for x in top_sellers) / SELLERS_LIMIT)

        return {
            'sellers': kpi_top_sellers,
        }

    def _build_trend_graph(self, module, cumulative=False):
        """
        Returns a line graph with the purchase amount of a module for the last 3 months.

        @param module: a loempia.module
        @param cumulative: do we start from the first day (3 month before today) or not (1 month before abscisse)
        """

        points_limit = 10

        d0 = date.today() - relativedelta(days=RECENT_THRESHOLD)
        d1 = date.today() + relativedelta(days=1)

        # Every days from 1 month ago to today
        ticks = [d0 + relativedelta(days=+i) for i in range((d1 - d0).days)]
        pruned_ticks = self._get_pruned_tick_values(ticks, points_limit)

        trend_graph = []
        for t in pruned_ticks:
            request.cr.execute("""
                SELECT SUM(p.price) AS price
                FROM
                    loempia_module_purchase AS p
                WHERE
                    p.state not in ('draft', 'cancel') AND
                    p.referrer_module_id in %(referrer_module_ids)s AND
                    p.date_order BETWEEN date %(start_date)s AND date %(end_date)s
            """, {
                'referrer_module_ids': tuple(module.template_id.module_ids.ids),
                'start_date': d0.strftime(DEFAULT_SERVER_DATE_FORMAT) if cumulative else (t - relativedelta(months=+1)).strftime(DEFAULT_SERVER_DATE_FORMAT),
                'end_date': t.strftime(DEFAULT_SERVER_DATE_FORMAT)
            })

            sales = request.cr.fetchall()
            trend_graph.append([
                t.strftime(DEFAULT_SERVER_DATE_FORMAT),
                sales[0][0] or 0
            ])

        return trend_graph

    def _build_sales_monthly_graph(self, module_type, category_id=None, user_id=None):
        """
        Returns a graph with
            x : months (12)
            y: sales of all modules, filtered by
                * module_type ('apps' or 'themes') (referrer_module_id.is_theme) EXCEPT if user_id
                * category_id (referrer_module_id.category_id)
                * user_id (module_maintainer_id)
        """
        # TODO: take timezones into account if we end up setting timeszones on new users
        # (probably won't happen soon since we create the users with auth_oauth)
        first_day = (date.today() - relativedelta(months=11)).replace(day=1).strftime(DEFAULT_SERVER_DATE_FORMAT)
        request.env.cr.execute("""
            SELECT to_char(date_trunc('month', p.date_order::timestamp), 'Month YYYY') as month,sum(p.price) as price
            FROM loempia_module_purchase p
            JOIN loempia_module m
                ON m.id = p.referrer_module_id
            WHERE p.state not in ('draft','cancel')
                AND date_order >= %(first_day)s
                AND ((%(category_id)s IS NULL) OR (m.category_id = %(category_id)s))
                AND
                    CASE WHEN %(user_id)s IS NULL
                        THEN m.is_theme = %(is_theme)s
                        ELSE p.module_maintainer_id = %(user_id)s
                    END
            GROUP BY date_trunc('month', p.date_order::timestamp)
            ORDER BY date_trunc('month', p.date_order::timestamp);
        """, {
            'first_day': first_day,
            'is_theme': module_type == 'themes',
            'category_id': category_id,
            'user_id': user_id
        })
        purchases = request.env.cr.dictfetchall()
        return [[' '.join(p['month'].split()), p['price']] for p in purchases]

    def _get_pruned_tick_values(self, ticks, nb_desired_ticks):
        if nb_desired_ticks == 0:
            return ticks

        nb_values = len(ticks)
        keep_one_of = max(1, floor(nb_values / float(nb_desired_ticks)))

        pruned_ticks = []
        for i in range(len(ticks)):
            if i % keep_one_of == 0:
                pruned_ticks.append(ticks[i])

        return pruned_ticks

    def serve(self, module_type, path, module, can_comment, rst=False):
        pricelist = self._get_user_pricelist()
        purchased = request.env['loempia.module.purchase'].user_has_purchase(module)
        # avoid extra prefetching on modules found in user_has_purchase
        module = module[0]
        template = module.template_id
        if module.status not in ['approved', 'invalid'] and not (module.price > 0 and purchased):
            return request.not_found()

        # TODO RAR - legacy: will disappear whe we switch to 1 purchase per version
        latest_deps_ids = [
            m._get_commercial_series_module_id().id for m in module.all_depend_ids.filtered(lambda m: m.price_single > 0.0)
            if not m.repo_id.is_server and not m.repo_id.is_odoo_enterprise
        ]

        # related version will either contain itself if version >=13.0 or every version <=12.0.
        related_versions = template.get_packaged_version(module)

        # Dec 2018: Since the pricing of ENT and SH are now monthly based on the apps
        # sold, we need to be more upfront with the dependencies as if an app depends
        # on a community module, it will install the latter, hence upselling the
        # contract of the client
        # To solve this, we now find the 'upselling' dependencies and display them as
        # such on the apps page. To retrieve this information we ask accounts to filter out the free dependencies.
        enterprise_apps_per_version = set(module.series_id.enterprise_apps.mapped('name'))
        module_dependencies = set(module.all_depend_ids.mapped('name'))
        enterprise_apps_deps = list(enterprise_apps_per_version & module_dependencies)

        maintenance_tooltip = ''
        if module._can_maintenance_be_charged():
            default_batch_size = int(request.env['ir.config_parameter'].sudo().get_param('loempia.lines_batch_size', '100'))
            price_per_batch = request.env['ir.qweb.field.monetary'].value_to_html(pricelist.price_per_batch, options={
                'display_currency': pricelist.currency_id,
            })
            estimated_maintenance_cost = request.env['ir.qweb.field.monetary'].value_to_html(module.maintenance_cost, options={
                'display_currency': pricelist.currency_id,
            })

            # RAR: The tooltip for maintenance code is computed in python since we cannot invoke t-field to properly display the price in its currency.
            maintenance_tooltip = "Odoo's public price to maintain custom code and third party module\
                is <b>%(price_per_batch)s</b> per <b>%(default_batch_size)s</b> lines of code. The service includes unlimited technical support\
                tickets, bug fixes and upgrade. <br> This module has a total of <b>%(number_of_lines)s</b> lines of code (dependencies included).\
                The estimated maintenance cost for this module is then <b>%(estimated_maintenance_cost)s</b> per month." % {
                'default_batch_size': default_batch_size,
                'number_of_lines': module.total_lines_of_code,
                'price_per_batch': price_per_batch,
                'estimated_maintenance_cost': estimated_maintenance_cost,
            }

        # Deactivated atm until we define which error should be inputed to the vendor
        invalid_mods = module[0].status == 'invalid' or module[0].all_depend_ids.filtered(lambda m: m.status == 'invalid')

        purchase_user = module.get_purchases_users()
        ratings_confirmed = {rating: rating.author_id in purchase_user for rating in module.ratings_ids}

        return request.render('loempia.apps', {
            'app_html': module[0]["description_html"], #if rst else file_content,
            'module': module[0],
            'main_object': module,
            'module_type': module_type,
            'application_dependencies': [dep for dep in module.all_depend_ids if dep.name in enterprise_apps_deps],
            'bundled_dependencies': [dep for dep in module.all_depend_ids if not dep.repo_id.is_server and not dep.repo_id.is_odoo_enterprise],
            'ratings_confirmed': ratings_confirmed,
            'purchased': purchased,
            'currency_id': pricelist.currency_id,
            'can_comment': can_comment,
            'rst': rst,
            'need_navbar': True,
            'is_app_page': True,
            'related_versions': related_versions,
            'latest_module_ids': [module._get_commercial_series_module_id().id] + latest_deps_ids,
            'maintenance_tooltip': maintenance_tooltip,
            'seo_module': module.template_id.latest_module_id,
            # 'invalid_mods': invalid_mods,
            'no_header': request.context.get('no_header', False),
            'no_footer': request.context.get('no_footer', False),
        })

    def serve_static(self, path):
        with open(path, 'rb') as file_:
            import mimetypes
            mime, encoding = mimetypes.guess_type(path)

            # There is no reason to serve any static file except
            # images and videos that would be included in the module
            # description. This has been abused in the past.
            if not mime or not mime.startswith(("image/", "video/")):
                raise NotFound()

            # Refuse inappropriate destinations, for safety reasons
            # (svg files allowed here)
            check_sec_fetch_dest(["image", "video"])

            if mime:
                headers = [('Content-Type', mime)]
            return request.make_response(file_.read(), headers)

    ###########################
    # Loading images from url #
    ###########################
    @route('/apps/module_image', auth='public')
    def module_image(self, image_id):
        check_sec_fetch_dest(["image"])

        res = request.env['ir.attachment'].sudo().search_read([
            ('id', '=', int(image_id)),
            ('res_model', '=', 'loempia.module'),
            ('mimetype', '=like', 'image/%')
        ], ["datas", "name"])
        if res:
            res = res[0]
        else:
            raise NotFound()
        image = io.BytesIO(base64.standard_b64decode(res["datas"]))
        return http.send_file(image, filename=res['name'])

    @route('/apps/module_image_cropped', auth='public')
    def module_image_cropped(self, image_id):
        check_sec_fetch_dest(["image"])

        res = request.env['ir.attachment'].sudo().search_read([
            ('id', '=', int(image_id)),
            ('res_model', '=', 'loempia.module'),
            ('mimetype', '=like', 'image/%')
        ], ["datas", "name"])
        if res:
            res = res[0]
        else:
            raise NotFound()
        image_cropped = image_process(res['datas'], crop='center', size=(190, 95))
        image = io.BytesIO(base64.b64decode(image_cropped))
        return http.send_file(image, filename=res['name'])

    ##########################
    # Loading icons from url #
    ##########################
    @route('/apps/icon_image', auth='public')
    def icon_image(self, module_id):
        # `icon_image` is supposed to be a safe image because the computed field
        # opens it with Image.open before returning it.
        # We can check Sec-Fetch-Dest anyway, for extra safety.
        check_sec_fetch_dest(["image"])

        # TODO use path and take on disc
        # we should get rid of this controller and use standard code as much as possible.
        # This means adapting every call to icon_image in all mail templates for instance
        res = request.env['loempia.module'].sudo().search_read([('id', '=', int(module_id))], ["icon_image"])
        if res:
            res = res[0]
        else:
            raise NotFound()
        image = io.BytesIO(base64.standard_b64decode(res["icon_image"]))
        return http.send_file(image, filename=module_id + 'png')

class LoempiaDisputesController(http.Controller):

    @http.route([
        '/apps/support/<string:so_ref>/<string:module_name>',
        '/apps/support/<string:so_ref>/<string:module_name>/<int:module_id>',
        '/apps/support/<string:so_ref_1>/<string:so_ref_2>/<string:module_name>',
        '/apps/support/<string:so_ref_1>/<string:so_ref_2>/<string:module_name>/<int:module_id>',
    ], type='http', auth="user", website=True)
    def claim_page(self, module_name, so_ref=None, so_ref_1=None, so_ref_2=None, module_id=None, access_token='', **kwargs):
        order_name = so_ref or (so_ref_1 + '/' + so_ref_2)
        purchase = request.env['loempia.module.purchase'].get_claim_purchase(
            [('order_name', '=', order_name), ('state', '!=', 'draft'), ('module_id.name', '=', module_name)],
            module_id=module_id,
            access_token=access_token,
        ).sudo()
        if not purchase:
            return request.render("loempia.claim_page_no_access", {
                'error_message': 'This account is not the one that is linked to purchase %s.' % order_name,
            })
        module = purchase.module_id
        render_ctx = {'rates': False}
        existing_rating = request.env['loempia.module.rating']
        if module_id:
            existing_rating = existing_rating.search([
                ('author_id', '=', request.env.user.partner_id.id),
                ('rating', '>', 0),
                ('module_id', '=', module_id)
            ])

        if kwargs.get('rating'):
            render_ctx['rating'] = kwargs.get('rating', 0)
            render_ctx['comment'] = kwargs.get('comment', '')
            render_ctx['subject'] = kwargs.get('subject', '')
            render_ctx['module_id'] = module_id
            render_ctx.update({'rates': True, 'new_rates': True})
        elif existing_rating:
            body = html2plaintext(existing_rating.body)
            render_ctx.update({
                'comment': body,
                'subject': existing_rating.subject,
                'rating': existing_rating.rating,
                'module_id': module_id,
                'rates': True,
                'new_rates': False,
            })

        max_claim_delay = int(request.env['ir.config_parameter'].sudo().get_param('loempia.max_claim_delay', '60'))
        remaining_days = max_claim_delay - (fields.datetime.now() - purchase.date_order).days
        user_messages = purchase.message_ids.filtered(lambda m: m.author_id.id == purchase.user_id.partner_id.id)
        first_contact_date = user_messages and user_messages[-1].create_date
        can_request_intervention = request.env['res.users'].has_group('loempia.group_manager') or (
            remaining_days > 0
            and not purchase.intervention_requested
            and purchase.state not in ['draft', 'cancel']
            and purchase.message_ids
            and (first_contact_date and (fields.datetime.now() - first_contact_date).days >= 15)
        )

        can_request_refund = not purchase.refund_id and (remaining_days > 0\
            and request.env.user == purchase.module_maintainer_id\
            or request.env['res.users'].has_group('loempia.group_functional')
        )

        module_type = 'modules'
        if module.is_theme:
            module_type = 'themes'
        elif module.category_id.name in CATEGORIES_INDUSTRY:
            module_type = 'industries'

        render_ctx.update({
            'purchased': True,
            'purchase': purchase,
            'module': module,
            'module_type': module_type,
            'can_request_intervention': can_request_intervention,
            'can_request_refund': can_request_refund,
            'remaining_days': remaining_days,
            'currency_id': request.env['res.company'].browse(1).currency_id,
            'support_page': True,
            'seo_module': module.template_id.latest_module_id,
            'reasons': request.env['loempia.refund.reason'].sudo().search([]),
            'error': 'err' in kwargs,
            'access_token': access_token,
        })
        return request.render("loempia.claim_page", render_ctx)

    @http.route([
        '/apps/support/<int:module_id>',
    ], type='http', auth="user", website=True)
    def fuzzy_support(self, module_id, **kwargs):
        module = request.env['loempia.module'].sudo().browse(module_id)._get_commercial_series_module_id()
        purchase = request.env['loempia.module.purchase'].search([
            ('module_id', '=', module.id),
            ('user_id', '=', request.env.user.id),
            ('state', 'not in', ('draft', 'cancel', 'exception')),
        ], limit=1, order='date_order desc').sudo()
        if not purchase:
            module = request.env['loempia.module'].sudo().browse(module_id)
            return request.render("loempia.claim_page_no_access", {
                'error_message': 'This account has not purchased module %s for Odoo %s.' % (module.name, module.series_id.name),
            })
        else:
            return self.claim_page(module.name, so_ref=purchase.order_name, module_id=module_id, **kwargs)

    @http.route('/apps/support/<int:order_id>/<int:module_id>/report', type='http', auth="user", website=True, methods=['POST'])
    def report_dispute(self, order_id=None, module_id=None, **post):
        access_token = post.get('access_token', '')
        url_args = {'access_token': access_token}

        purchase = request.env['loempia.module.purchase'].get_claim_purchase(
            [('order_id', '=', order_id), ('state', '=', 'confirmed')],
            module_id=module_id,
            access_token=access_token,
        )

        max_claim_delay = int(request.env['ir.config_parameter'].sudo().get_param('loempia.max_claim_delay', '60'))
        remaining_days = max_claim_delay - (fields.datetime.now() - purchase.date_order).days
        user_messages = purchase.message_ids.filtered(lambda m: m.author_id.id == purchase.user_id.partner_id.id)
        first_contact_date = user_messages and user_messages[-1].create_date
        can_request_intervention = request.env['res.users'].has_group('loempia.group_manager') or (
            remaining_days > 0
            and not purchase.intervention_requested
            and purchase.state not in ['draft', 'cancel', 'exception']
            and purchase.message_ids
            and (first_contact_date and (fields.datetime.now() - first_contact_date).days >= 15)
        )
        if not purchase.exists() or not can_request_intervention:
            return request.render("website.page_404")

        if not purchase.intervention_requested:
            purchase = purchase.sudo()

            # create support ticket
            params = {
                'order_name': purchase.order_name,
                'module_name': purchase.module_id.name,
                'module_id': purchase.module_id.id
            }
            ticket_url = '{PROD}/openerp_apps/support_ticket'.format(PROD=BUYAPPS_HOST)

            result = False
            try:
                req = requests.post(ticket_url, data=params, auth=SaasAuth(), timeout=TIMEOUT)
                req.raise_for_status()
                result = req.json()
            except Exception as e:
                logger.info('Failed request to %s: error %s', ticket_url, e)
                url_args['error'] = '1'

            if result:
                purchase.intervention_requested = True  # nobody can push on the button anymore
                apps_bot = request.env.ref("loempia.res_partner_odoo_bot").sudo()
                if not apps_bot.exists():
                    apps_bot = request.env['res.users'].browse(SU).sudo().partner_id
                message = purchase.message_log(
                            body='%s requested the intervention of Odoo. You will hear from us shortly.' % (request.env.user.name,),
                            subject=False,
                            author=apps_bot,
                            message_type='comment',
                            # subtype=False
                        )

                if request.env.user.partner_id.id not in purchase.message_follower_ids.mapped('partner_id').ids:
                    purchase.message_subscribe([request.env.user.partner_id.id])

                subject = 'Odoo intervention requested for %s' % purchase.order_name
                purchase.notify_followers(apps_bot, 'intervention', subject)

                url_args['fragment'] = 'loempia-comments'

        url = purchase._build_claim_url(**url_args)
        return request.redirect(url)

    def _refund_request(self, purchase, reason, comment, other=None):
        if reason == 'other':
            reason_text = other
        else:
            reason = request.env['loempia.refund.reason'].sudo().browse(int(reason))
            reason_text = reason.reason
        max_claim_delay = int(request.env['ir.config_parameter'].sudo().get_param('loempia.max_claim_delay', '60'))
        remaining_days = max_claim_delay - (fields.datetime.now() - purchase.date_order).days
        can_process_refund = remaining_days > 0\
            and request.env.user == purchase.module_maintainer_id\
            or request.env['res.users'].has_group('loempia.group_functional')

        if not reason_text or not can_process_refund:
            return request.render("website.page_404")
        # A tmp access does not grant the access to the _refund_request
        url = purchase.sudo()._build_claim_url()

        refund = request.env['loempia.refund'].sudo().create({
            'purchase_id': purchase.id,
            'maintainer_message': comment,
            'instigator_id': request.env.user.id,
            'reason': reason_text,
        })
        purchase.sudo().refund_id = refund.id
        apps_bot = request.env.ref("loempia.res_partner_odoo_bot").sudo()
        if not apps_bot.exists():
            apps_bot = request.env['res.users'].browse(SU).sudo().partner_id
        body = 'The customer can apply for a refund.'

        purchase.message_log(
            body=body,
            subject=False,
            author=apps_bot,
            message_type='comment',
        )

        followers = purchase.sudo().message_follower_ids.mapped('partner_id')
        if request.env.user.partner_id not in followers:
            purchase.message_subscribe([request.env.user.partner_id.id])

        return request.redirect(url)

    def _refund_validation(self, purchase, comment):
        url_args = {}
        purchase = purchase.sudo()
        if (request.env.user != purchase.user_id and not request.env.user.has_group('loempia.group_functional'))\
            or not purchase.refund_id or purchase.refund_id.expired:
            return request.render("website.page_404")

        # create support ticket
        params = {
            'module_id': purchase.module_id.id,
            'module_maintainer_name': purchase.module_maintainer_id.name,
            'module_name': purchase.module_id.name,
            'order_id': purchase.order_id,
            'order_name': purchase.order_name,
            'customer_name': purchase.user_id.sudo().name,
            'origin_name': request.env.user.name,
            'reason': purchase.refund_id.reason
        }

        ticket_url = '{PROD}/openerp_apps/refund_ticket'.format(PROD=BUYAPPS_HOST)
        result = False
        try:
            result = self._create_support_ticket(ticket_url, params)
        except Exception as e:
            logger.info('Failed request to %s: error %s', ticket_url, e)
            url_args['error'] = '1'

        if result:
            purchase.intervention_requested = True  # nobody can push on the button anymore
            apps_bot = request.env.ref("loempia.res_partner_odoo_bot").sudo()
            if not apps_bot.exists():
                apps_bot = request.env['res.partner'].sudo().search([('user_ids', 'in', [1])]).sudo()
            if purchase.module_maintainer_id == purchase.refund_id.instigator_id:
                body = 'The publisher agreed to refund the customer. Odoo has already been contacted and will issue the refund shortly.'
            else:
                body = 'Odoo will issue the refund shortly.'

            purchase.message_log(
                body=body,
                subject=False,
                author=apps_bot,
                message_type='comment',
            )
            purchase.refund_id.write({'user_message': comment, 'state': 'accepted'})

            if request.env.user.partner_id.id not in purchase.message_follower_ids.mapped('partner_id').ids:
                purchase.message_subscribe([request.env.user.partner_id.id])

            url_args['fragment'] = 'loempia-comments'
            # For now, we do not notify the users when a refund is requested
            # purchase.notify_followers(apps_bot, '')

        url = purchase._build_claim_url(**url_args)
        return request.redirect(url)

    def _create_support_ticket(self, endpoint, params):
        req = requests.post(endpoint, data=params, auth=SaasAuth(), timeout=TIMEOUT)
        req.raise_for_status()
        return req.json()

    @http.route('/apps/support/<int:refund_id>/refund', type='http', auth="user", website=True)
    def extend_refund(self, refund_id):
        refund = request.env["loempia.refund"].sudo().browse(int(refund_id)).exists()
        if not refund or not refund.expired or not request.env.user.has_group('loempia.group_functional'):
            return request.render("website.page_404")
        refund._extend_limit_date()
        purchase = refund.purchase_id

        purchase._message_log(
            body="The odoo support accept to increase the limit date for the refund request!",
            subject=False,
        )

        url = purchase._build_claim_url(fragment='loempia-comments')
        return request.redirect(url)

    @http.route('/apps/support/<int:order_id>/<int:module_id>/refund', type='http', auth="user", website=True, methods=['POST'])
    def refund(self, order_id, module_id, comment, reason=None, other=''):
        comment = html2plaintext(comment)
        other = html2plaintext(other)

        purchase = request.env['loempia.module.purchase'].get_claim_purchase([('order_id', '=', order_id), ('state', '=', 'confirmed')], module_id=module_id)
        if not purchase or not comment\
            or (not purchase.refund_id and not reason)\
            or (purchase.refund_id and purchase.sudo().refund_id.state != 'pending'):
            return request.render("website.page_404")
        if purchase.refund_id:
            return self._refund_validation(purchase, comment)
        else:
            return self._refund_request(purchase, reason, comment, other)

    @route('/apps/support/message', type='http', auth="user", website=True, methods=['POST'])
    def post_message_claim(self, purchase_id=False, **post):
        access_token = post.get('access_token', '')
        try:
            purchase_id = int(purchase_id)
        except (ValueError):
            return request.redirect('/apps')
        else:
            user = request.env.user
            # sudo() is here to access the partner on the user_id
            purchase = request.env['loempia.module.purchase'].get_claim_purchase([('id', '=', purchase_id)], access_token=access_token).sudo()

            if not purchase.exists() or (purchase.state != 'confirmed' and not request.env.user.has_group('loempia.group_functional')):
                return request.redirect('/apps')

            url = purchase._build_claim_url(access_token=access_token, fragment='loempia-comments')

            if 'message' not in post or len(post['message']) == 0:
                return request.redirect(url)

            message = purchase.message_log(
                body=post['message'],
                subject=False,
                author=user.partner_id,
                message_type='comment',
            )
            # add the author as follower of in the thread if not part of it
            purchase.message_subscribe([purchase.module_maintainer_id.partner_id.id, purchase.user_id.partner_id.id])
            if user.partner_id.id not in purchase.message_follower_ids.mapped('partner_id').ids:
                purchase.message_subscribe([user.partner_id.id])

            purchase.notify_followers(user.partner_id)

            return request.redirect(url)


class LoempiaDirectOAuthLogin(OAuthLogin):
    @route()
    def web_login(self, *args, **kw):
        response = super(LoempiaDirectOAuthLogin, self).web_login(*args, **kw)
        if 'nooauth' not in kw and 'error' not in response.qcontext:
            odoo_provider_id = request.env.ref('auth_oauth.provider_openerp').id
            odoo_provider = next((p for p in response.qcontext.get('providers', []) if p['id'] == odoo_provider_id), None)
            if odoo_provider:
                response = request.redirect(odoo_provider['auth_link'], local=False)
        return response

class LoempiaSignup(AuthSignupHome):
    @route()
    def web_auth_signup(self, *args, **kw):
        # Non-Internal users should never be created from apps directly.
        # No need for fancy redirection in this case, since there is nothing that points
        # to the signup page of the apps store. This is just to handle the rare (but not
        # non-existant) cases were people explicitly reach the signup page of the apps store
        # directly by explicitly entering its URL.
        signup_url = "{PROD}/web/signup?redirect=/apps".format(PROD=BUYAPPS_HOST)
        return request.redirect(signup_url, local=False)
